/****************************************************************************/
/*! @file
@brief DirectShow

obt@w_O
-----------------------------------------------------------------------------
	Copyright (C) 2004 T.Imoto
-----------------------------------------------------------------------------
@author		T.Imoto
@date		2004/08/05
@note
			2004/08/05	T.Imoto		
*****************************************************************************/

#include <streams.h>
#include "BufferRenderer.h"

#ifdef _DEBUG
#include <stdio.h>
#include "DShowException.h"
#endif

//----------------------------------------------------------------------------
//##	TBufferRenderer
//----------------------------------------------------------------------------
//! @brief	  	This goes in the factory template table to create new filter instances
//! @param		pUnk : W IUnknown C^[tFCXւ̃|C^B
//! @param		phr : \bh̐Es HRESULT l󂯎ϐւ̃|C^B
//----------------------------------------------------------------------------
CUnknown * WINAPI TBufferRenderer::CreateInstance( LPUNKNOWN pUnk, HRESULT *phr )
{
	CUnknown	*punk = new TBufferRenderer( NAME("Buffer Renderer"), pUnk, phr );
	if( punk == NULL )
		*phr = E_OUTOFMEMORY;
	return punk;
}
#pragma warning(disable: 4355)	// RXgN^̃x[Xothisgƃ[jOôł}~
//----------------------------------------------------------------------------
//! @brief	  	TBufferRenderer constructor
//! @param		pName : fobÔ߂ɎgpLqւ̃|C^B
//! @param		pUnk : WꂽL҃IuWFNgւ̃|C^B
//! @param		phr : HRESULT lւ̃|C^B
//----------------------------------------------------------------------------
TBufferRenderer::TBufferRenderer( TCHAR *pName, LPUNKNOWN pUnk, HRESULT *phr )
 : CBaseVideoRenderer( CLSID_BufferRenderer, pName, pUnk, phr )
, m_InputPin( this, &m_InterfaceLock, phr, L"Input" )
, m_Allocator( this, NAME("Allocator"), GetOwner(), phr )
{
	//CBaseRender::m_pInputPinɃ|C^ݒ肷B
	m_pInputPin = &m_InputPin;

	// Store and AddRef the texture for our use.
	*phr = S_OK;
	m_Buffer[0] = NULL;
	m_Buffer[1] = NULL;

	m_IsBufferOwner[0] = false;
	m_IsBufferOwner[1] = false;

	m_FrontBuffer = 0;

	m_StartFrame = 0;
}
#pragma warning(default: 4355)
//----------------------------------------------------------------------------
//! @brief	  	TBufferRenderer destructor
//----------------------------------------------------------------------------
TBufferRenderer::~TBufferRenderer()
{
	//CBaseRender::m_pInputPinɃ|C^ZbgB
	//ȂCBaseRender̃fXgN^deleteĂ܂̂ŒӁI
    m_pInputPin = NULL;

	// ŊmۂĂꍇobt@̉
	FreeFrontBuffer();
	FreeBackBuffer();
}
//----------------------------------------------------------------------------
//! @brief	  	vꂽC^[tFCXԂ
//! 
//! Overriden to say what interfaces we support and where
//! @param		riid : C^[tFCXIID
//! @param		ppv : C^[tFCXԂ|C^[ւ̃|C^
//! @return		G[R[h
//----------------------------------------------------------------------------
STDMETHODIMP TBufferRenderer::NonDelegatingQueryInterface( REFIID riid, void **ppv )
{
	CheckPointer( ppv, E_POINTER );
	if(riid == IID_IRendererBufferAccess) {
		*ppv = static_cast<IRendererBufferAccess*>(this);
		static_cast<IUnknown*>(*ppv)->AddRef();
		return S_OK;
	} else if(riid == IID_IRendererBufferVideo) {
		*ppv = static_cast<IRendererBufferVideo*>(this);
		static_cast<IUnknown*>(*ppv)->AddRef();
		return S_OK;
	}
	return CBaseVideoRenderer::NonDelegatingQueryInterface( riid, ppv );
}
//----------------------------------------------------------------------------
//! @brief	  	̃fBA ^CvtB^󂯓邩ǂmF
//! 
//! This method forces the graph to give us an R8G8B8 video type, making our copy 
//! to texture memory trivial.
//! @param		pmt : ĂꂽfBA ^Cv܂ CMediaType IuWFNgւ̃|C^
//! @return		ĂꂽfBA ^Cv󂯓Ȃ S_OK ԂB
//!				łȂ S_FALSE G[ R[hԂB
//----------------------------------------------------------------------------
HRESULT TBufferRenderer::CheckMediaType( const CMediaType *pmt )
{
	HRESULT		hr = E_FAIL;
	VIDEOINFO	*pvi;

	// Reject the connection if this is not a video type
	if( *pmt->FormatType() != FORMAT_VideoInfo )
		return E_INVALIDARG;

	// Only accept RGB32
	pvi = (VIDEOINFO *)pmt->Format();
	if( IsEqualGUID( *pmt->Type(), MEDIATYPE_Video) )
	{
		if( IsEqualGUID( *pmt->Subtype(), MEDIASUBTYPE_RGB32) || 
			IsEqualGUID( *pmt->Subtype(), MEDIASUBTYPE_ARGB32) )
		{
			hr = S_OK;
			m_MtIn = (*pmt);
		}
	}

	return hr;
}
//----------------------------------------------------------------------------
//! @brief	  	Graph connection has been made. 
//! @param		pmt : fBA ^Cvw肷 CMediaType IuWFNgւ̃|C^
//! @return		G[R[h
//----------------------------------------------------------------------------
HRESULT TBufferRenderer::SetMediaType( const CMediaType *pmt )
{
	CAutoLock cAutoLock(&m_BufferLock);	// NeBJZNV

	// Retrive the size of this media type
	VIDEOINFO *pviBmp;						// Bitmap info header
	pviBmp = (VIDEOINFO *)pmt->Format();
	m_VideoWidth  = pviBmp->bmiHeader.biWidth;
	m_VideoHeight = abs(pviBmp->bmiHeader.biHeight);
	m_VideoPitch = m_VideoWidth * 4;	// RGB32Ɍߑł

	if( !IsAllocatedFrontBuffer() )
		AllocFrontBuffer( GetBufferSize() );

	if( !IsAllocatedBackBuffer() )
		AllocBackBuffer( GetBufferSize() );

	return S_OK;
}
//----------------------------------------------------------------------------
//! @brief	  	A sample has been delivered. Copy it to the texture.
//! @param		pSample : Tv IMediaSample C^[tFCXւ̃|C^
//! @return		G[R[h
//----------------------------------------------------------------------------
HRESULT TBufferRenderer::DoRenderSample( IMediaSample * pSample )
{
	DWORD	*pBmpBuffer, *pTxtBuffer;	// Bitmap buffer, texture buffer
	BYTE	*pTxtOrgPos;

//	if( m_bEOS ) return S_OK;

	CAutoLock cAutoLock(&m_BufferLock);	// NeBJZNV

	// Get the video bitmap buffer
	pSample->GetPointer( reinterpret_cast<BYTE**>(&pBmpBuffer) );

	// Get the texture buffer & pitch
	pTxtBuffer = reinterpret_cast<DWORD*>(GetBackBuffer());
	pTxtOrgPos = reinterpret_cast<BYTE*>(pTxtBuffer);

	HRESULT		hr;
	LONG		EventParam1 = -1;
	LONGLONG	TimeStart = 0;
	LONGLONG	TimeEnd = 0;

	if( SUCCEEDED(hr = pSample->GetMediaTime( &TimeStart, &TimeEnd )) )
	{
		EventParam1 = (LONG)TimeStart;
	}
	if( m_StopFrame && EventParam1 >= m_StopFrame )
		return S_OK;	// ĐȂt[

	if( pTxtBuffer == pBmpBuffer )	// ÕAP[^[gĂ
	{
		SwapBuffer( pSample );	// FrontBackobt@ւ
		if( m_pSink )
			m_pSink->Notify( EC_UPDATE, EventParam1, NULL );
		return S_OK;
	}

	// ÕAP[^[ł͂Ȃ̂ŃRs[
#if 0
	// ɃRs[(㉺])
	{
		int		height = m_VideoHeight;
		int		width = m_VideoWidth;
		pBmpBuffer += width * (height-1);
		for( int j = 0; j < height; j++ )
		{
			for( int i = 0; i < width; i++ )
			{
				pTxtBuffer[i] = pBmpBuffer[i];
			}
			pBmpBuffer -= width;
			pTxtBuffer += width;
		}
	}
#else
	// ォ牺ɃRs[
	{
		int		height = m_VideoHeight;
		int		width = m_VideoWidth;
		for( int j = 0; j < height; j++ )
		{
			for( int i = 0; i < width; i++ )
			{
				pTxtBuffer[i] = pBmpBuffer[i];
			}
			pBmpBuffer += width;
			pTxtBuffer += width;
		}
	}
#endif
	if( m_pSink )
		m_pSink->Notify( EC_UPDATE, EventParam1, NULL );
	SwapBuffer( pSample );	// FrontBackobt@ւ
	return S_OK;
}
//---------------------------------------------------------------------------
//! @brief	  	tgobt@ƃobNobt@ւ
//! @param		pSample : TvB̒̃|C^ύX
//----------------------------------------------------------------------------
void TBufferRenderer::SwapBuffer( IMediaSample *pSample )
{
	CAutoLock cAutoLock(&m_BufferLock);	// NeBJZNV
	if( m_FrontBuffer == 1 )
	{
		SetPointer( pSample, m_Buffer[1] );
		m_FrontBuffer = 0;
	}
	else
	{
		SetPointer( pSample, m_Buffer[0] );
		m_FrontBuffer = 1;
	}
}
//---------------------------------------------------------------------------
//! @brief	  	tgobt@Ƀ蓖Ă
//! @param		size : 蓖ĂTCY
//----------------------------------------------------------------------------
void TBufferRenderer::AllocFrontBuffer( size_t size )
{
	CAutoLock cAutoLock(&m_BufferLock);	// NeBJZNV
	BYTE	*buff = NULL;

	FreeFrontBuffer();
	if( m_FrontBuffer == 1 )
	{
		buff = m_Buffer[1] = reinterpret_cast<BYTE*>(CoTaskMemAlloc(size));
		m_IsBufferOwner[1] = true;
	}
	else
	{
		buff = m_Buffer[0] = reinterpret_cast<BYTE*>(CoTaskMemAlloc(size));
		m_IsBufferOwner[0] = true;
	}

	if( buff == NULL )
		throw L"Cannot allocate memory in filter.";
}
//---------------------------------------------------------------------------
//! @brief	  	obNobt@Ƀ蓖ĂB
//! @param		size : 蓖ĂTCY
//----------------------------------------------------------------------------
void TBufferRenderer::AllocBackBuffer( size_t size )
{
	CAutoLock cAutoLock(&m_BufferLock);	// NeBJZNV
	BYTE	*buff = NULL;

	FreeBackBuffer();
	if( m_FrontBuffer == 1 )
	{
		buff = m_Buffer[0] = reinterpret_cast<BYTE*>(CoTaskMemAlloc(size));
		m_IsBufferOwner[0] = true;
	}
	else
	{
		buff = m_Buffer[1] = reinterpret_cast<BYTE*>(CoTaskMemAlloc(size));
		m_IsBufferOwner[1] = true;
	}

	if( buff == NULL )
		throw L"Cannot allocate memory in filter.";
}
//---------------------------------------------------------------------------
//! @brief	  	tgobt@Ɋ蓖ĂĂ郁J
//!
//! A蓖ĂĂ郁ÃNXɂĊ蓖Ăꂽ̂łȂꍇ́A
//! ȂB
//----------------------------------------------------------------------------
void TBufferRenderer::FreeFrontBuffer()
{
	CAutoLock cAutoLock(&m_BufferLock);	// NeBJZNV
	if( m_FrontBuffer == 1 )
	{
		if( m_Buffer[1] != NULL )
		{
			if( m_IsBufferOwner[1] )
				CoTaskMemFree( m_Buffer[1] );
			m_Buffer[1] = NULL;
		}
		m_IsBufferOwner[1] = false;
	}
	else
	{
		if( m_Buffer[0] != NULL )
		{
			if( m_IsBufferOwner[0] )
				CoTaskMemFree( m_Buffer[0] );
			m_Buffer[0] = NULL;
		}
		m_IsBufferOwner[0] = false;
	}
}
//---------------------------------------------------------------------------
//! @brief	  	obNobt@Ɋ蓖ĂĂ郁J
//!
//! A蓖ĂĂ郁ÃNXɂĊ蓖Ăꂽ̂łȂꍇ́A
//! ȂB
//----------------------------------------------------------------------------
void TBufferRenderer::FreeBackBuffer()
{
	CAutoLock cAutoLock(&m_BufferLock);	// NeBJZNV
	if( m_FrontBuffer == 1 )
	{
		if( m_Buffer[0] != NULL )
		{
			if( m_IsBufferOwner[0] )
				CoTaskMemFree( m_Buffer[0] );
			m_Buffer[0] = NULL;
		}
		m_IsBufferOwner[0] = false;
	}
	else
	{
		if( m_Buffer[1] != NULL )
		{
			if( m_IsBufferOwner[1] )
				CoTaskMemFree( m_Buffer[1] );
			m_Buffer[1] = NULL;
		}
		m_IsBufferOwner[1] = false;
	}
}
//---------------------------------------------------------------------------
//! @brief	  	tgobt@Ƀobt@ւ̃|C^ݒ肷
//! @param		buff : obt@ւ̃|C^
//----------------------------------------------------------------------------
void TBufferRenderer::SetFrontBuffer( BYTE *buff )
{
	CAutoLock cAutoLock(&m_BufferLock);	// NeBJZNV
	FreeFrontBuffer();
	if( m_FrontBuffer == 1 )
		m_Buffer[1] = buff;
	else
		m_Buffer[0] = buff;
}
//---------------------------------------------------------------------------
//! @brief	  	obNobt@Ƀobt@ւ̃|C^ݒ肷
//! @param		buff : obt@ւ̃|C^
//----------------------------------------------------------------------------
void TBufferRenderer::SetBackBuffer( BYTE *buff )
{
	CAutoLock cAutoLock(&m_BufferLock);	// NeBJZNV
	FreeBackBuffer();
	if( m_FrontBuffer == 1 )
		m_Buffer[0] = buff;
	else
		m_Buffer[1] = buff;
	
	SetPointer( buff );
}
//---------------------------------------------------------------------------
//! @brief	  	tgobt@ւ̃|C^擾
//! @return		obt@ւ̃|C^
//----------------------------------------------------------------------------
BYTE *TBufferRenderer::GetFrontBuffer()
{
	CAutoLock cAutoLock(&m_BufferLock);	// NeBJZNV
	if( m_FrontBuffer == 1 )
		return m_Buffer[1];
	else
		return m_Buffer[0];
}
//---------------------------------------------------------------------------
//! @brief	  	obNobt@ւ̃|C^擾
//! @return		obt@ւ̃|C^
//----------------------------------------------------------------------------
BYTE *TBufferRenderer::GetBackBuffer()
{
	CAutoLock cAutoLock(&m_BufferLock);	// NeBJZNV
	if( m_FrontBuffer == 1 )
		return m_Buffer[0];
	else
		return m_Buffer[1];
}
//----------------------------------------------------------------------------
//! @brief	  	tgobt@ݒ肵܂B
//! @param		buff : tgobt@pobt@ւ̃|C^
//! @param		size : obt@̃TCYnϐւ̃|C^B@n
//!					buffNULL̎Aɗ~TCYԂ
//! @return		G[R[h
//----------------------------------------------------------------------------
HRESULT TBufferRenderer::SetFrontBuffer( BYTE *buff, long *size )
{
	if( m_State == State_Running )
		return S_FALSE;

	CAutoLock cAutoLock(&m_BufferLock);	// NeBJZNV
	if( buff == NULL && size != NULL )
	{
		*size = GetBufferSize();
		return S_OK;
	}
	if( buff == NULL || size == NULL )
		return E_POINTER;

	if( (*size) != GetBufferSize() )
		return E_INVALIDARG;

	FreeFrontBuffer();
	SetFrontBuffer(buff);
	return S_OK;
}
//----------------------------------------------------------------------------
//! @brief	  	obNobt@ݒ肵܂B
//! @param		buff : obNobt@pobt@ւ̃|C^
//! @param		size : obt@̃TCYnϐւ̃|C^B@n
//!					buffNULL̎Aɗ~TCYԂ
//! @return		G[R[h
//----------------------------------------------------------------------------
HRESULT TBufferRenderer::SetBackBuffer( BYTE *buff, long *size )
{
	if( m_State == State_Running )
		return S_FALSE;

	CAutoLock cAutoLock(&m_BufferLock);	// NeBJZNV
	if( buff == NULL && size != NULL )
	{
		*size = GetBufferSize();
		return S_OK;
	}
	if( buff == NULL || size == NULL )
		return E_POINTER;

	if( (*size) != GetBufferSize() )
		return E_INVALIDARG;

	FreeBackBuffer();
	SetBackBuffer(buff);
	return S_OK;
}
//----------------------------------------------------------------------------
//! @brief		tgobt@ւ̃|C^擾܂B
//! @param		buff : tgobt@ւ̃|C^Ԃ߂̃obt@ւ̃|C^
//! @param		size : obt@̃TCYԂϐւ̃|C^
//! @return		G[R[h
//----------------------------------------------------------------------------
HRESULT TBufferRenderer::GetFrontBuffer( BYTE **buff, long *size )
{
	CAutoLock cAutoLock(&m_BufferLock);	// NeBJZNV
	*buff = GetFrontBuffer();
	*size = GetBufferSize();
	return S_OK;
}
//----------------------------------------------------------------------------
//! @brief	  	obNobt@ւ̃|C^擾܂B
//! @param		buff : obNobt@ւ̃|C^Ԃ߂̃obt@ւ̃|C^
//! @param		size : obt@̃TCYԂϐւ̃|C^
//! @return		G[R[h
//----------------------------------------------------------------------------
HRESULT TBufferRenderer::GetBackBuffer( BYTE **buff, long *size )
{
	CAutoLock cAutoLock(&m_BufferLock);	// NeBJZNV
	*buff = GetBackBuffer();
	*size = GetBufferSize();
	return S_OK;
}
//----------------------------------------------------------------------------
//! @brief	  	1t[̕ϕ\Ԃ擾܂
//! @param		pAvgTimePerFrame : 1t[̕ϕ\
//! @return		G[R[h
//----------------------------------------------------------------------------
HRESULT TBufferRenderer::get_AvgTimePerFrame( REFTIME *pAvgTimePerFrame )
{
	if( pAvgTimePerFrame ) {
		*pAvgTimePerFrame = (reinterpret_cast<VIDEOINFOHEADER *>(m_MtIn.Format())->AvgTimePerFrame) / 10000000.0;
		return S_OK;
	}
	else
		return E_POINTER;
}
//----------------------------------------------------------------------------
//! @brief	  	rfI̕擾܂
//! @param		pVideoWidth : rfI̕
//! @return		G[R[h
//----------------------------------------------------------------------------
HRESULT TBufferRenderer::get_VideoWidth( long *pVideoWidth )
{
	if( pVideoWidth ) {
		*pVideoWidth = reinterpret_cast<VIDEOINFOHEADER *>(m_MtIn.Format())->bmiHeader.biWidth;
		return S_OK;
	}
	else
		return E_POINTER;
}
//----------------------------------------------------------------------------
//! @brief	  	rfI̍擾܂
//! @param		pVideoHeight : rfI̍
//! @return		G[R[h
//----------------------------------------------------------------------------
HRESULT TBufferRenderer::get_VideoHeight( long *pVideoHeight )
{
	if( pVideoHeight ) {
		*pVideoHeight = reinterpret_cast<VIDEOINFOHEADER *>(m_MtIn.Format())->bmiHeader.biHeight;
		return S_OK;
	}
	else
		return E_POINTER;
}
//----------------------------------------------------------------------------
//! @brief	  	Xg[~OJnꂽɃR[
//!
//! Jnt[L^B
//! @return		G[R[h
//----------------------------------------------------------------------------
HRESULT TBufferRenderer::OnStartStreaming(void)
{
	HRESULT		hr;
	CComPtr<IMediaSeeking>	mediaSeeking;
	
	if( m_pGraph )
	{
		if( m_pGraph->QueryInterface( &mediaSeeking ) != S_OK )
			mediaSeeking = NULL;
	}

	bool		bGetTime = false;
	LONGLONG	Current = 0;
	if( mediaSeeking.p != NULL )
	{	// IMediaSeekinggĎԂ̎擾݂
		GUID	Format;
		if( SUCCEEDED(hr = mediaSeeking->GetTimeFormat( &Format ) ) )
		{
			if( SUCCEEDED(hr = mediaSeeking->GetCurrentPosition( &Current )) )
			{
				if( IsEqualGUID( TIME_FORMAT_MEDIA_TIME, Format ) )
				{
					double	renderTime = Current / 10000000.0;
					REFTIME	AvgTimePerFrame;	// REFTIME :  b\{x_B
					if( SUCCEEDED( hr = get_AvgTimePerFrame( &AvgTimePerFrame ) ) )
					{
						m_StartFrame = (LONG)(renderTime / AvgTimePerFrame + 0.5);
						bGetTime = true;
					}
				}
				else if( IsEqualGUID( TIME_FORMAT_FRAME, Format ) )
				{
					m_StartFrame = (LONG)Current;
					bGetTime = true;
				}
			}
		}
	}
	if( bGetTime == false )
		m_StartFrame = 0;
	
	mediaSeeking.Release();

	return CBaseVideoRenderer::OnStartStreaming();
}
//----------------------------------------------------------------------------
//! @brief	  	_OOɃR[
//!
//! fBATvɃfBA^CL^B
//! fBA^C͊Jnt[Ɍ݂̃Xg[ԂẐɂȂB
//! AtB^IMediaSeekingC^[tFCXpłȂꍇ́A
//! ̃_[tB^`悵t[ƃhbvt[ZB
//! ̏ꍇAʂ̃tB^Ńhbvt[͂킩Ȃ̂ŁA
//! ኱xB
//! @param		pMediaSample : fBATv
//----------------------------------------------------------------------------
void TBufferRenderer::OnRenderStart( IMediaSample *pMediaSample )
{
	CBaseVideoRenderer::OnRenderStart(pMediaSample);

	HRESULT		hr;
	bool		bGetTime = false;
	LONGLONG	Current = 0, Stop = 0;
	IMediaSeeking	*mediaSeeking = NULL;
	if( GetMediaPositionInterface( IID_IMediaSeeking, (void**)&mediaSeeking) == S_OK )
	{
		GUID	Format;
		if( SUCCEEDED(hr = mediaSeeking->GetTimeFormat( &Format ) ) )
		{
			if( SUCCEEDED(hr = mediaSeeking->GetCurrentPosition( &Current )) &&
				SUCCEEDED(hr = mediaSeeking->GetStopPosition( &Stop )) )
			{
				if( IsEqualGUID( TIME_FORMAT_MEDIA_TIME, Format ) )
				{
					double	renderTime = Current / 10000000.0;
					double	stopTime = Stop / 10000000.0;
					REFTIME	AvgTimePerFrame;	// REFTIME :  b\{x_B
					if( SUCCEEDED( hr = get_AvgTimePerFrame( &AvgTimePerFrame ) ) )
					{
						Current = (LONGLONG)(renderTime / AvgTimePerFrame + 0.5);
						Stop = (LONGLONG)(stopTime / AvgTimePerFrame + 0.5);
						bGetTime = true;
					}
				}
				else if( IsEqualGUID( TIME_FORMAT_FRAME, Format ) )
				{
					bGetTime = true;
				}
			}
		}
		mediaSeeking->Release();
		mediaSeeking = NULL;
	}
	LONGLONG	TimeStart = m_StartFrame + m_cFramesDrawn + m_cFramesDropped;;
	LONGLONG	TimeEnd = m_StartFrame + m_cFramesDrawn + m_cFramesDropped;;
	if( bGetTime == true )
	{
		TimeStart = m_StartFrame + Current;
		TimeEnd = m_StartFrame + Current;
		m_StopFrame = m_StartFrame + static_cast<LONG>(Stop);
	}
	else
	{
		TimeStart = m_StartFrame + m_cFramesDrawn + m_cFramesDropped;;
		TimeEnd = m_StartFrame + m_cFramesDrawn + m_cFramesDropped;;
		m_StopFrame = 0;
	}
	pMediaSample->SetMediaTime( &TimeStart, &TimeEnd );
}
//----------------------------------------------------------------------------
//##	TBufferRendererInputPin
//----------------------------------------------------------------------------
//! @brief	  	̓sIuWFNg\z܂B
//! @param		pRenderer : _[IuWFNgw肵܂B
//! @param		pInterfaceLock : CCritSec bNւ̃|C^ŁAԈڍsp邽߂ɎgpB@n
//!					̓tB^ bN CBaseFilter.m_pLock Ɠl̃NeBJ ZNVƂȂ肤B 
//! @param		phr : \bh̐Es HRESULT l擾ϐ̃|C^B
//! @param		name : IuWFNg̃fobOp̖O镶B
//----------------------------------------------------------------------------
TBufferRendererInputPin::TBufferRendererInputPin( TBufferRenderer *pRenderer, CCritSec *pInterfaceLock, HRESULT *phr, LPCWSTR name)
 : CRendererInputPin( pRenderer, phr, name ), m_pRenderer( pRenderer ), m_pInterfaceLock( pInterfaceLock ),
 m_ActiveAllocator( false )
{
}
//----------------------------------------------------------------------------
//! @brief	  	fXgN^B݂͉ȂB
//----------------------------------------------------------------------------
TBufferRendererInputPin::~TBufferRendererInputPin()
{
}
//----------------------------------------------------------------------------
//! @brief	  	ÕAP[^Lǂׂ܂B
//! @return		LȂTRUEԂ܂B
//----------------------------------------------------------------------------
bool TBufferRendererInputPin::ActiveAllocator( void ) const
{
	return m_ActiveAllocator;
}
//----------------------------------------------------------------------------
//! @brief	  	ÕAP[^IuWFNg蓖Ă܂B
//! @param		ppAllocator : ԂAP[^[
//----------------------------------------------------------------------------
STDMETHODIMP TBufferRendererInputPin::GetAllocator( IMemAllocator **ppAllocator )
{
	CAutoLock cInterfaceLock(m_pInterfaceLock);
	CheckPointer(ppAllocator,E_POINTER);

	// AP[^܂ݒ肳ĂȂƂ
	if (m_pAllocator == NULL) {
		m_pAllocator = &(m_pRenderer->m_Allocator);
		m_pAllocator->AddRef();
	}
	// QƃJEgĉ̓C^tF[X̎dlłB
	m_pAllocator->AddRef();
	*ppAllocator = m_pAllocator;

	return S_OK;
}
//----------------------------------------------------------------------------
//! @brief	  	AP[^܂ƂɌĂяo܂B
//! @param		pAllocator ̐ڑŎgpAP[^w肵܂B
//! @param		bReadOnly ̃AP[^̃Tvǂ݂ƂpȂTRUEw肵܂B
//! @return		G[R[h
//----------------------------------------------------------------------------
STDMETHODIMP TBufferRendererInputPin::NotifyAllocator( IMemAllocator * pAllocator, BOOL bReadOnly )
{
	CAutoLock cInterfaceLock(m_pInterfaceLock);

	// NXĂяo
	HRESULT hr = CBaseInputPin::NotifyAllocator(pAllocator, bReadOnly);
	if( FAILED(hr) )
		return hr;

	//ÕAP[^LǂL^܂
	m_ActiveAllocator = (pAllocator == (&(m_pRenderer->m_Allocator)));

	return S_OK;
}
//----------------------------------------------------------------------------
//! @brief	  	w肵fBATvɃ|C^ݒ肵܂
//! @param		media : fBATv
//! @param		ptr : ݒ肷|C^
//----------------------------------------------------------------------------
void TBufferRendererInputPin::SetPointer( IMediaSample *media, BYTE *ptr )
{
	m_pRenderer->m_Allocator.SetPointer( media, ptr );
}
//----------------------------------------------------------------------------
//! @brief	  	AP[^[fBATvɃ|C^ݒ肵܂
//! @param		ptr : ݒ肷|C^
//----------------------------------------------------------------------------
void TBufferRendererInputPin::SetPointer( BYTE *ptr )
{
	m_pRenderer->m_Allocator.SetPointer( ptr );
}
//----------------------------------------------------------------------------
//## TBufferRendererAllocator
//----------------------------------------------------------------------------
//! @brief	  	RXgN^
//! @param		pRenderer : _[IuWFNgw肵܂B
//! @param		pName : IuWFNg̃fobOp̖O镶B
//! @param		pUnk : WꂽL҃IuWFNgւ̃|C^B
//! @param		phr : \bh̐Es HRESULT l擾ϐ̃|C^B
//----------------------------------------------------------------------------
TBufferRendererAllocator::TBufferRendererAllocator( TBufferRenderer *pRenderer, TCHAR *pName, LPUNKNOWN pUnk, HRESULT *phr)
 : CBaseAllocator(pName, pUnk, phr), m_pMediaSample(NULL), m_pRenderer(pRenderer)
{}
//----------------------------------------------------------------------------
//! @brief	  	CBaseAllocator::DecommitR[Ȃ΂ȂȂ̂ŁAR[B
//----------------------------------------------------------------------------
TBufferRendererAllocator::~TBufferRendererAllocator()
{
	Decommit();
	ASSERT(m_lAllocated == m_lFree.GetCount());
	// Free up all the CMediaSamples
	CMediaSample *pSample;
	for(;;)
	{
		pSample = m_lFree.RemoveHead();
		if( pSample != NULL )
			delete pSample;
		else
			break;
	}
	m_lAllocated = 0;
}
//----------------------------------------------------------------------------
//! @brief	  	ȂBAP[^[邱Ƃ͂ȂB
//----------------------------------------------------------------------------
void TBufferRendererAllocator::Free( void )
{
}
//----------------------------------------------------------------------------
//! @brief	  	蓖ĂāAXgɒǉ
//! @return		G[R[h
//----------------------------------------------------------------------------
HRESULT TBufferRendererAllocator::Alloc( void )
{
	CAutoLock	lck(this);

	/* Check he has called SetProperties */
	HRESULT hr = CBaseAllocator::Alloc();
	if( FAILED(hr) )
		return hr;

	/* If the requirements haven't changed then don't reallocate */
	if( hr == S_FALSE )
		return NOERROR;

	ASSERT(hr == S_OK); // we use this fact in the loop below
	LPBYTE	pBuffer = static_cast<LPBYTE>(m_pRenderer->GetBackBuffer());

	CMediaSample *pSample;
	pSample = new CMediaSample( NAME("buffer media sample"), this, &hr, pBuffer, m_lSize );
	ASSERT(SUCCEEDED(hr));
	if( pSample == NULL ) {
		return E_OUTOFMEMORY;
	}
	m_lFree.Add(pSample);
	m_lAllocated++;
	m_pMediaSample = pSample;
	return NOERROR;
}

//----------------------------------------------------------------------------
//! @brief	  	v郁̏ڍׂݒ肷B
//! @param		pRequest : obt@v܂ ALLOCATOR_PROPERTIES \̂̃|C^
//! @param		pActual : ۂ̃obt@ vpeB󂯎 ALLOCATOR_PROPERTIES \̂̃|C^
//! @return		G[R[h
//----------------------------------------------------------------------------
STDMETHODIMP TBufferRendererAllocator::SetProperties( ALLOCATOR_PROPERTIES* pRequest, ALLOCATOR_PROPERTIES* pActual )
{
	CheckPointer( pActual, E_POINTER );
	ValidateReadWritePtr( pActual, sizeof(ALLOCATOR_PROPERTIES) );
	CAutoLock	cObjectLock(this);

    ZeroMemory( pActual, sizeof(ALLOCATOR_PROPERTIES) );

	if( m_bCommitted == TRUE)
		return VFW_E_ALREADY_COMMITTED;

	if( m_lFree.GetCount() < m_lAllocated )	// m_lAllocated1ȏ̎̓G[ɂ悢H
		return VFW_E_BUFFERS_OUTSTANDING;

	if( pRequest->cBuffers == 1 && pRequest->cbBuffer == m_pRenderer->GetBufferSize() &&
		pRequest->cbAlign == 1 && pRequest->cbPrefix == 0 )
	{
		*pActual = *pRequest;
		m_lSize = pRequest->cbBuffer;
		m_lCount = pRequest->cBuffers;
		m_lAlignment = pRequest->cbAlign;
		m_lPrefix = pRequest->cbPrefix;
		m_bChanged = TRUE;
		return S_OK;
	}

	return VFW_E_BADALIGN;
}
//----------------------------------------------------------------------------
//! @brief	  	w肵fBATvɃ|C^ݒ肵܂
//! @param		media : fBATv
//! @param		ptr : ݒ肷|C^
//----------------------------------------------------------------------------
void TBufferRendererAllocator::SetPointer( IMediaSample *media, BYTE *ptr )
{
	BYTE	*pBufferParam, *pBufferOwn;
	if( media )
	{
		media->GetPointer( &pBufferParam);
		if( m_pMediaSample != NULL )
		{
			m_pMediaSample->GetPointer( &pBufferOwn );
			if( pBufferOwn == pBufferParam )	// obt@wĂ̂ŁAێĂTvƓƌȂ
			{
				LONG	cBytes = m_pMediaSample->GetSize();	// TCY͕ςĂȂƌȂAOɃ`FbNĂ
				m_pMediaSample->SetPointer( ptr, cBytes );
			}
		}
	}
}
